// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// Red-Black Balanced Binary Tree Algorithm adapted from:
// Thomas Niemann <thomasn@epaperpress.com>
// See also:
// Thomas H. Cormen, et al
// Introduction to Algorithms
// (MIT Electrical Engineering and Computer Science)
// (C) 1990 MIT Press
//

type
  ESymTreeError = class (Exception);

  PSymRBTreeNode = ^TSymRBTreeNode;
  TSymRBTreeNode = record
    fIndex: Integer;
    fLeft: PSymRBTreeNode;
    fRight: PSymRBTreeNode;
    fParent: PSymRBTreeNode;
    fColor: Byte;
    fStrInd: Integer;
    fCellInd: Integer;
  end;

  TSymbolTreeIterator = procedure (handle: Integer; strInd, cellInd: Integer);

  TSymbolTree = class
  protected
    fNodesUsed: Integer;
    fRoot: PSymRBTreeNode;
    fSentinel: PSymRBTreeNode;
    fPool: array of PSymRBTreeNode;
    fPoolUsed: Integer;
    fFreeList: PSymRBTreeNode;

    function GetStrings (handle: Integer): Integer;
    function GetCells (handle: Integer): Integer;
    procedure SetCells (handle, value: Integer);
    function AllocNode (): PSymRBTreeNode;
    procedure FreeNode (node: PSymRBTreeNode);
    function GetMaxHandle (): Integer;

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Clear ();

    function Append (strInd, cellInd: Integer): Integer; // insert string, return handle
    procedure Delete (handle: Integer);
    function Find (strInd: Integer): Integer; // -1: no such string

    procedure Iterate (IterProc: TSymbolTreeIterator);

    function ValidHandle (handle: Integer): Boolean;

    property Count: Integer read fNodesUsed;
    property MaxHandle: Integer read GetMaxHandle;
    property Strings [handle: Integer]: Integer read GetStrings;
    property Cells [handle: Integer]: Integer read GetCells write SetCells;
  end;


{ helper procedures }
procedure SymRotateLeft (tree: TSymbolTree; x: PSymRBTreeNode);
// rotate node x to left
var
  y: PSymRBTreeNode;
begin
  y := x.fRight;

  (* establish x.Right link *)
  x.fRight := y.fLeft;
  if y.fLeft <> tree.fSentinel then y.fLeft.fParent := x;

  (* establish y.Parent link *)
  if y <> tree.fSentinel then y.fParent := x.fParent;
  if x.fParent <> nil then
  begin
    if x = x.fParent.fLeft then x.fParent.fLeft := y else x.fParent.fRight := y;
  end
  else tree.fRoot := y;

  (* link x and y *)
  y.fLeft := x;
  if x <> tree.fSentinel then x.fParent := y;
end;

procedure SymRotateRight (tree: TSymbolTree; x: PSymRBTreeNode);
// rotate node x to right
var
  y: PSymRBTreeNode;
begin
  y := x.fLeft;

  (* establish x.Left link *)
  x.fLeft := y.fRight;
  if y.fRight <> tree.fSentinel then y.fRight.fParent := x;

  (* establish y.Parent link *)
  if y <> tree.fSentinel then y.fParent := x.fParent;
  if x.fParent <> nil then
  begin
    if x = x.fParent.fRight then x.fParent.fRight := y else x.fParent.fLeft := y;
  end
  else tree.fRoot := y;

  (* link x and y *)
  y.fRight := x;
  if x <> tree.fSentinel then x.fParent := y;
end;

procedure SymInsertFixup (tree: TSymbolTree; x: PSymRBTreeNode);
//  maintain red-black tree balance after inserting node x
var
  y: PSymRBTreeNode;
begin
  (* check red-black properties *)
  while (x <> tree.fRoot) and (x.fParent.fColor = RBT_RED) do
  begin
    (* we have a violation *)
    if x.fParent = x.fParent.fParent.fLeft then
    begin
      y := x.fParent.fParent.fRight;
      if y.fColor = RBT_RED then
      begin
        (* uncle is red *)
        x.fParent.fColor := RBT_BLACK;
        y.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        x := x.fParent.fParent;
      end
      else
      begin
        (* uncle is black *)
        if x = x.fParent.fRight then
        begin
          (* make x a fLeft child *)
          x := x.fParent;
          SymRotateLeft(tree, x);
        end;

        (* recolor and rotate *)
        x.fParent.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        SymRotateRight(tree, x.fParent.fParent);
      end;
    end
    else
    begin
      (* mirror image of above code *)
      y := x.fParent.fParent.fLeft;
      if y.fColor = RBT_RED then
      begin
        (* uncle is red *)
        x.fParent.fColor := RBT_BLACK;
        y.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        x := x.fParent.fParent;
      end
      else
      begin
        (* uncle is black *)
        if x = x.fParent.fLeft then
        begin
          x := x.fParent;
          SymRotateRight(tree, x);
        end;
        x.fParent.fColor := RBT_BLACK;
        x.fParent.fParent.fColor := RBT_RED;
        SymRotateLeft(tree, x.fParent.fParent);
      end;
    end;
  end;

  tree.fRoot.fColor := RBT_BLACK;
end;

procedure SymDeleteFixup (tree: TSymbolTree; x: PSymRBTreeNode);
// maintain red-black tree balance after deleting node x
var
  w: PSymRBTreeNode;
begin
  while (x <> tree.fRoot) and (x.fColor = RBT_BLACK) do
  begin
    if x = x.fParent.fLeft then
    begin
      w := x.fParent.fRight;
      if w.fColor = RBT_RED then
      begin
        w.fColor := RBT_BLACK;
        x.fParent.fColor := RBT_RED;
        SymRotateLeft(tree, x.fParent);
        w := x.fParent.fRight;
      end;
      if (w.fLeft.fColor = RBT_BLACK) and (w.fRight.fColor = RBT_BLACK) then
      begin
        w.fColor := RBT_RED;
        x := x.fParent;
      end
      else
      begin
        if w.fRight.fColor = RBT_BLACK then
        begin
          w.fLeft.fColor := RBT_BLACK;
          w.fColor := RBT_RED;
          SymRotateRight(tree, w);
          w := x.fParent.fRight;
        end;
        w.fColor := x.fParent.fColor;
        x.fParent.fColor := RBT_BLACK;
        w.fRight.fColor := RBT_BLACK;
        SymRotateLeft(tree, x.fParent);
        x := tree.fRoot;
      end
    end
    else
    begin
      w := x.fParent.fLeft;
      if w.fColor = RBT_RED then
      begin
        w.fColor := RBT_BLACK;
        x.fParent.fColor := RBT_RED;
        SymRotateRight(tree, x.fParent);
        w := x.fParent.fLeft;
      end;
      if (w.fRight.fColor = RBT_BLACK) and (w.fLeft.fColor = RBT_BLACK) then
      begin
        w.fColor := RBT_RED;
        x := x.fParent;
      end
      else
      begin
        if w.fLeft.fColor = RBT_BLACK then
        begin
          w.fRight.fColor := RBT_BLACK;
          w.fColor := RBT_RED;
          SymRotateLeft(tree, w);
          w := x.fParent.fLeft;
        end;
        w.fColor := x.fParent.fColor;
        x.fParent.fColor := RBT_BLACK;
        w.fLeft.fColor := RBT_BLACK;
        SymRotateRight(tree, x.fParent);
        x := tree.fRoot;
      end;
    end;
  end;

  x.fColor := RBT_BLACK;
end;


{ TSymbolTree }
constructor TSymbolTree.Create ();
begin
  SetLength(fPool, 0);
  Clear();
end;

destructor TSymbolTree.Destroy ();
begin
  Clear();
  Dispose(fPool[0]); // sentinel
  SetLength(fPool, 0);
end;

procedure TSymbolTree.Clear ();
var
  f: Integer;
begin
  f := fPoolUsed;
  while f > 0 do begin Dec(f); Dispose(fPool[f]); end;
  SetLength(fPool, 0);

  fSentinel := nil;
  fPoolUsed := 0; fNodesUsed := 0; fFreeList := nil;
  fSentinel := AllocNode();
  fSentinel.fLeft := fSentinel; fSentinel.fRight := fSentinel; fSentinel.fParent := nil;
  fSentinel.fColor := RBT_BLACK;
  fRoot := fSentinel;
end;

function TSymbolTree.GetStrings (handle: Integer): Integer;
begin
  if (handle < 0) or (handle >= fPoolUsed) or (fPool[handle].fColor = RBT_UNUSED) then
    raise ESymTreeError.Create('invalid symbol index (GetStrings)');
  result := fPool[handle].fStrInd;
end;

function TSymbolTree.GetCells (handle: Integer): Integer;
begin
  if (handle < 0) or (handle >= fPoolUsed) or (fPool[handle].fColor = RBT_UNUSED) then
    raise ESymTreeError.Create('invalid symbol index (GetCells)');
  result := fPool[handle].fCellInd;
end;

procedure TSymbolTree.SetCells (handle, value: Integer);
begin
  if (handle < 0) or (handle >= fPoolUsed) or (fPool[handle].fColor = RBT_UNUSED) then
    raise ESymTreeError.Create('invalid symbol index (SetCells)');
  fPool[handle].fCellInd := value;
end;

procedure TSymbolTree.FreeNode (node: PSymRBTreeNode);
begin
  if (node = nil) or (node = fSentinel) then exit;
  ASSERT(node.fColor <> RBT_UNUSED);
  node.fLeft := nil; node.fRight := nil; node.fColor := RBT_UNUSED;
  node.fParent := fFreeList;
  fFreeList := node;
end;

function TSymbolTree.AllocNode (): PSymRBTreeNode;
var
  f: Integer;
begin
  if fFreeList <> nil then
  begin
    result := fFreeList;
    fFreeList := fFreeList.fParent;
  end
  else
  begin
    ASSERT(fPoolUsed <= Length(fPool));
    if fPoolUsed = Length(fPool) then
    begin
      SetLength(fPool, Length(fPool)+GROW_DELTA);
      for f := fPoolUsed to High(fPool) do fPool[f] := nil;
    end;
    New(result); fPool[fPoolUsed] := result;
    result.fIndex := fPoolUsed; Inc(fPoolUsed);
  end;
  result.fLeft := fSentinel; result.fRight := fSentinel; result.fParent := nil;
  result.fStrInd := 0;
  result.fCellInd := 0;
  result.fColor := RBT_RED;
end;

function TSymbolTree.Find (strInd: Integer): Integer;
// -1: no such string
var
  r: Integer;
  node: PSymRBTreeNode;
begin
  result := -1; node := fRoot;
  while node <> fSentinel do
  begin
    r := strInd-node.fStrInd;
    if r = 0 then break;
    if r < 0 then node := node.fLeft else node := node.fRight;
  end;

  if node <> fSentinel then result := node.fIndex;
end;

function TSymbolTree.Append (strInd, cellInd: Integer): Integer;
// insert string, return handle
var
  current, parent, newNode: PSymRBTreeNode;
  r: Integer;
begin
  (* find where node belongs *)
  current := fRoot; parent := nil;
  while current <> fSentinel do
  begin
    r := strInd-current.fStrInd;
    if r = 0 then
    begin
      // duplicate found
      result := current.fIndex;
      exit;
    end;
    parent := current;
    if r < 0 then current := current.fLeft else current := current.fRight;
  end;

  (* setup new node *)
  newNode := AllocNode();
  newNode.fParent := parent;
  newNode.fLeft := fSentinel;
  newNode.fRight := fSentinel;
  newNode.fColor := RBT_RED;
  newNode.fStrInd := strInd;
  newNode.fCellInd := cellInd;

  (* insert node in tree *)
  if parent <> nil then
  begin
    if strInd-parent.fStrInd < 0 then parent.fLeft := newNode
    else parent.fRight := newNode;
  end
  else fRoot := newNode;
  SymInsertFixup(self, newNode);

  Inc(fNodesUsed);
  result := newNode.fIndex;
end;

procedure TSymbolTree.Delete (handle: Integer);
var
  x, y, z: PSymRBTreeNode;
  color: Byte;
begin
  if handle = 0 then exit;

  if (handle < 0) or (handle >= fPoolUsed) or (fPool[handle].fColor = RBT_UNUSED) then
    raise ESymTreeError.Create('invalid symbol index (Delete)');

  (* find node in tree *)
  z := fPool[handle];
  ASSERT(z <> fSentinel);

  if (z.fLeft = fSentinel) or (z.fRight = fSentinel) then
  begin
    (* y has a nil node as a child *)
    y := z;
  end
  else
  begin
    (* find tree successor with a sentinel node as a child *)
    y := z.fRight;
    while y.fLeft <> fSentinel do y := y.fLeft;
  end;

  (* x is y's only child *)
  if y.fLeft <> fSentinel then x := y.fLeft else x := y.fRight;

  (* remove y from the parent chain *)
  x.fParent := y.fParent;
  if y.fParent <> nil then
  begin
    if y = y.fParent.fLeft then y.fParent.fLeft := x else y.fParent.fRight := x;
  end
  else fRoot := x;

  color := y.fColor;

  if y <> z then
  begin
    (* swap y and z *)
    y.fLeft := z.fLeft;
    y.fRight := z.fRight;
    y.fParent := z.fParent;
    if z.fParent <> nil then
    begin
      if z.fParent.fLeft = z then z.fParent.fLeft := y else z.fParent.fRight := y;
    end
    else fRoot := y;

    y.fRight.fParent := y;
    y.fLeft.fParent := y;

    y.fColor := z.fColor;
  end;

  if color = RBT_BLACK then SymDeleteFixup(self, x);

  FreeNode(z);
  Dec(fNodesUsed);
end;

procedure TSymbolTree.Iterate (IterProc: TSymbolTreeIterator);
const
  STACK_SIZE = 50;
var
  stack: array [0..STACK_SIZE-1] of PSymRBTreeNode;
  sp: Integer;
  node: PSymRBTreeNode;
begin
  sp := 0; node := fRoot;
  repeat
    while node <> fSentinel do
    begin
      if sp >= STACK_SIZE then raise Exception.Create('out of stack in TSymbolTree.Iterate()');
      stack[sp] := node; Inc(sp);
      node := node.fLeft;
    end;
    if sp = 0 then break;

    Dec(sp); node := stack[sp];
    IterProc(node.fIndex, node.fStrInd, node.fCellInd);
    node := node.fRight;
  until false;
end;

function TSymbolTree.GetMaxHandle (): Integer;
begin
  result := fPoolUsed-1;
end;

function TSymbolTree.ValidHandle (handle: Integer): Boolean;
begin
  result := (handle >= 0) and (handle < fPoolUsed) and (fPool[handle].fColor <> RBT_UNUSED);
end;
