(setq listtoolz *t*)

(defun safe-car (a)
  (if (null? a) *nil*
    (atom? a) a
    (car a)
  )
)

(defun safe-cdr (a)
  (if (null? a) *nil*
    (atom? a) *nil*
    (cdr a)
  )
)

(defun caar (x) (car (car x)))
(defun cadr (x) (car (cdr x)))
(defun cdar (x) (cdr (car x)))
(defun cddr (x) (cdr (cdr x)))
(defun caaar (x) (car (car (car x))))
(defun caadr (x) (car (car (cdr x))))
(defun cadar (x) (car (cdr (car x))))
(defun caddr (x) (car (cdr (cdr x))))
(defun cdaar (x) (cdr (car (car x))))
(defun cdadr (x) (cdr (car (cdr x))))
(defun cddar (x) (cdr (cdr (car x))))
(defun cdddr (x) (cdr (cdr (cdr x))))
(defun caaaar (x) (car (car (car (car x)))))
(defun caaadr (x) (car (car (car (cdr x)))))
(defun caadar (x) (car (car (cdr (car x)))))
(defun caaddr (x) (car (car (cdr (cdr x)))))
(defun cadaar (x) (car (cdr (car (car x)))))
(defun cadadr (x) (car (cdr (car (cdr x)))))
(defun caddar (x) (car (cdr (cdr (car x)))))
(defun cadddr (x) (car (cdr (cdr (cdr x)))))
(defun cdaaar (x) (cdr (car (car (car x)))))
(defun cdaadr (x) (cdr (car (car (cdr x)))))
(defun cdadar (x) (cdr (car (cdr (car x)))))
(defun cdaddr (x) (cdr (car (cdr (cdr x)))))
(defun cddaar (x) (cdr (cdr (car (car x)))))
(defun cddadr (x) (cdr (cdr (car (cdr x)))))
(defun cdddar (x) (cdr (cdr (cdr (car x)))))
(defun cddddr (x) (cdr (cdr (cdr (cdr x)))))


; ---purpose---
; Verify that at least one member of a list has a given property.
; Return *T*, if a member with the given property exists and otherwise *F*.

; ---args---
; A - list
; P - predicate to test desired property

; ---see-also---
; every, some, member, equal

; ---example---
; (any '((a b c) x (d e f)) atom?)

; ---code---
(defun any (x p)
  (while (not-null? x)
    (if (p (car x))  (return *t*))
    (setq x (cdr x))
  )
  *f*
)

; ---purpose---
; Compare two lists (s-expressions, in fact).
; Return *T*, if both expressions print equal.

; ---args---
; A - expression
; B - expression

; ---see-also---
; nequal, member, subst, any

; ---example---
; (equal '(a b (c d (e)) f) '(a b (c d (e)) f))

; ---code---
(defun equal (a b)
  (while *t*
    (if (<> (atom? a) (atom? b))  (return *f*))
    (if (null? a)  (return (null? b))
        (null? b)  (return (null? a))
        (atom? a)  (return (eq a b))
        (not (equal (car a) (car b)))  (return *f*)
    )
    (setq a (cdr a)  b (cdr b))
  )
)

; ---purpose---
; Retrieve a binding from an association list.
; An association list is a list of pairs where the CAR part of each pair
; holds a key and the CDR part of the pair holds the value associated with
; that key:
; -
; ((key1.value1) ... (keyN.valueN))

; ---args---
; X - key of value to be found
; A - association list

; ---see-also---
; pair, sassoc, sublis, member, tree-assoc

; ---example---
; (assoc 3 '((1 . i) (2 . ii) (3 . iii) (4 . iv)))

; ---code---
(defun assoc (x a)
  (while (not-null? a)
    (if (equal x (caar a)) (return (car a)))
    (setq a (cdr a))
  )
  *nil*
)

; ---purpose---
; Verify that all members of a list have a given property.
; Return *T*, if all members have the requested property and
; *F*, if at least one member does not have the property.

; ---args---
; A - list
; P - predicate to test desired property

; ---see-also---
; any, some, member, equal

; ---example---
; (every '(5 7 12) number?)

; ---code---
(defun every (x p)
  (while (not-null? x)
    (if (not (p (car x))) (return *f*))
    (setq x (cdr x))
  )
  *t*
)

; ---purpose---
; Extract members from a list that match a given predicate.

; ---args---
; A - list
; P - predicate to match

; ---see-also---
; remove, some, every

; ---example---
; (filter '(abc 123 def 456 ghi) number?)

; ---code---
(defun filter (a p | res)
  (while (not-null? a)
    (if (p (car a)) (setq res (cons (car a) res)))
    (setq a (cdr a))
  )
  (reverse res)
)

; ---purpose---
; Check if an expression X is the head of an expression Y.

; ---args---
; X - potential head
; Y - expression

; ---see-also---
; tailp, caar, listp

; ---example---
; (headp '(a b c) '(a b c d e f))

; ---code---
(defun headp (x y)
  (while (not-null? x)
    (if (null? y) (return *f*)
        (not (equal (car x) (car y))) (return *f*)
    )
    (setq  x (cdr x)  y (cdr y))
  )
  *t*
)

; ---purpose---
; Insert a member into an ordered list.
; The predicate P must be asymmetrical and evaluate to T for each two
; subsequent members of A.

; ---args---
; X - member to insert
; A - ordered list
; P - predicate defining the order of A

; ---see-also---
; remove, sort, orderedp

; ---example---
; (insert 4 '(3 5) <)

; ---code---
(defun insert (x a p | n)
  ; find place to insert, reversing A to N
  (if
    (not
      (while (not-null? a)
        (if (p x (car a)) (break (setq n (cons x n))))
        (setq n (cons (car a) n)  a (cdr a))
      )
    )  (setq n (cons x n))
  )
  ; reverse & append the tail
  (while (not-null? a) (setq  n (cons (car a) n)  a (cdr a)))
  (reverse n)
)

; ---purpose---
; Extract the last member of a list.

; ---args---
; X - list

; ---see-also---
; caar, append, nth

; ---example---
; (last '(a b c d e f))

; ---code---
(defun last (x)
  (while (not-null? (cdr x)) (setq x (cdr x)))
  (car x)
)

; ---purpose---
; Check if an expression is a (proper) list.
; Proper lists are:
; -
; NIL, cons[X;NIL], cons[Y;cons[X;NIL]], etc.

; ---args---
; X - expression

; ---see-also---
; pair?, litatom?

; ---example---
; (list? '(a b c))

; ---code---
(defun list? (x)
  (while (land (not-null? x) (not-atom? x)) (setq x (cdr x)))
  (null? x)
)

; ---purpose---
; Check if an expression is a literal (non-NIL) atom.

; ---args---
; X - expression

; ---see-also---
; pair?, list?

; ---example---
; (litatom? 'xyz)

; ---code---
(defun litatom? (x) (land (atom? x) (not-null? x)))

; ---purpose---
; Check if an expression is a pair (not an atom).

; ---args---
; X - expression

; ---see-also---
; list?, pair, litatom?

; ---example---
; (pairp '(a.b))

; ---code---
(defun pair? (x) (not-atom? x))

; ---purpose---
; Map two lists into one using a given function.
; The function must take two arguments. If the
; lists have different lengths, only N pairs are
; mapped where N is the length of the shorter one
; of the lists.

; ---args---
; X - list
; Y - list
; F - function

; ---see-also---
; mapcar, maplis, reduce

; ---example---
; (map '(a b c) '(d e f) cons)

; ---code---
(defun map (x y f | n)
  (while (land (not-null? x) (not-null? y))
    (setq  n (cons (f (car x) (car y)) n)  x (cdr x)  y (cdr y))
  )
  (reverse n)
)

; ---purpose---
; Apply a function to each member of a list.

; ---args---
; A - list
; F - function to apply to each member

; ---see-also---
; map, maplis, reduce

; ---example---
; (mapcar '(1 2 3) neg)

; ---code---
(defun mapcar (a f | n)
  (while (not-null? a)
    (setq  n (cons (f (car a)) n)  a (cdr a))
  )
  (reverse n)
)

; ---purpose---
; Apply a function to each CDR of a list.

; ---args---
; A - list
; F - function to apply to each sublist

; ---see-also---
; map, mapcar, reduce

; ---example---
; (maplis '(1 2 3) reverse)

; ---code---
(defun maplis (a f | n)
  (while (not-null? a)
    (setq  n (cons (f a) n)  a (cdr a))
  )
  (reverse n)
)

; ---purpose---
; Check if a list has a member that is identical
; to a given symbol.

; ---args---
; X - symbol to find
; A - list

; ---see-also---
; member, union, intersection, unique, any

; ---example---
; (memb 'c '(a b c d e f))

; ---code---
(defun memb (x a)
  (while (not-null? a)
    (if (eq x (car a)) (return *t*))
    (setq a (cdr a))
  )
  *f*
)

; ---purpose---
; Check if a list has a member that is equal to a given S-expression.

; ---args---
; X - expression to find
; A - list

; ---see-also---
; memb, union, intersection, unique, any

; ---example---
; (member '(c d) '(a b (c d) e f))

(defun member (x a)
  (while (not-null? a)
    (if (equal x (car a)) (return *t*))
    (setq a (cdr a))
  )
  *f*
)

; ---purpose---
; Check if the members of a list are in a given monotonic order.

; ---args---
; X - list
; P - predicate defining order

; ---see-also---
; insert, sort, unsort

; ---example---
; (ordered? '(9 8 7 6 5) >)

; ---code---
(defun ordered? (x p)
  (while (land (not-null? x) (not-null? (cdr x)))
    (if (p (car x) (cadr x)) (setq x (cdr x))
      (return *f*)
    )
  )
  *t*
)

; ---purpose---
; Build an association list from two lists.

; ---args---
; X - list containing keys
; Y - list containing values

; ---see-also---
; assoc, sassoc, tree-assoc

; ---example---
; (pair '(1 2 3) '(i ii iii))

; ---code---
(defun pair (a b) (map a b cons))

/*
; ---purpose---
; Reduce a list. Combine the first member of the
; list with the reduced rest of the list using a
; given function.
; (C) Nils M Holm, modified by Ketmar

; ---args---
; A - list to reduce
; X - default member (replaces NIL)
; F - combining function

; ---see-also---
; map, mapcar, maplis

; ---example---
; (reduce '(1 2 3 4) 5 +)

; ---model---
; (defun reduce (a x f)
; (cond ((null a) x)
;   (t (f (car a) (reduce (cdr a) x f)))))

; ---code---
(defun reduce (a x f | red)
  (setq red (lambda (a b)
    (if (null? a)  b
        (null? b)  (red (cdr a) (f (car a) x))
      (red (cdr a) (f (car a) b))
    )
  ))
  (red (reverse a) *nil*)
)
*/

; ---purpose---
; Retrieve a binding from an association list, providing a
; default value. For a definition of association lists, see
; ASSOC.

; ---args---
; X - key to find
; A - association list to search
; C - constant function to apply if X not in A

; ---see-also---
; assoc, pair, sublis, tree-assoc

; ---example---
; (sassoc 'iv '((i . 1) (ii . 2) (iii . 3)) (lambda () *f*))
; (sassoc 'ii '((i . 1) (ii . 2) (iii . 3)) (lambda () *f*))

; ---code---
(defun sassoc (x a c)
  (while (not-null? a)
    (if (eq x (caar a)) (return (car a)))
    (setq a (cdr a))
  )
  (c)
)
