{*******************************************************************************
'  Routines for emulating the 128K Spectrum's AY-3-8912 sound generator
'
'  Author: James Bagg <chipmunk_uk_1@hotmail.com>
'
'  With minor optimisations and mods by
'      Chris Cowley <ccowley@grok.co.uk>
'
'  Translation to Delphi Object Pascal by
'      Jari Korhonen <jarit.korhonen@luukku.com>
'
'  Additional changings by
'     Christian Hackbart <chackbart@sqc.de>
'
'  Additional changings by
'     Ketmar // Vampire Avalon
'    stereo out, etc...
'
'  Copyright (C)1999-2000 Grok Developments Ltd and James Bagg
'  http://www.grok.co.uk/   http://www.chipmunks-corner.co.uk
'
'  This program is free software; you can redistribute it and/or
'  modify it under the terms of the GNU General Public License
'  as published by the Free Software Foundation; either version 2
'  of the License, or (at your option) any later version.
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
'
'*******************************************************************************}
unit dspAYEngineU;

interface


const
  ayMaxVolume = 31;


type
  TAYRegArray = array [0..15] of Integer;
  TVolTableArray = array [0..63] of Integer;
  TAY8912State = record
    sampleRate: Integer;
    registerLatch: Integer;
    regs: TAYRegArray;
    updateStep: Double;
    periodA: Integer;
    periodB: Integer;
    periodC: Integer;
    periodN: Integer;
    periodE: Integer;
    countA: Integer;
    countB: Integer;
    countC: Integer;
    countN: Integer;
    countE: Integer;
    volA: Integer;
    volB: Integer;
    volC: Integer;
    volE: Integer;
    envelopeA: Integer;
    envelopeB: Integer;
    envelopeC: Integer;
    outputA: Integer;
    outputB: Integer;
    outputC: Integer;
    outputN: Integer;
    countEnv: Integer;
    hold: Integer;
    alternate: Integer;
    attack: Integer;
    holding: Integer;
    volTable2: TVolTableArray;
  end;

  TAY8912 = class
  private
    fAYPSG: TAY8912State;
    ayOutNoise: Integer;
    volA: Integer;
    volB: Integer;
    volC: Integer;
    lOut0: Integer;
    lOut1: Integer;
    lOut2: Integer;
    ayLeft: Integer;
    ayNextEvent: Integer;

    fLeftChan: Integer;
    fRightChan: Integer;

    fChannelA: Integer;
    fChannelB: Integer;
    fChannelC: Integer;

    fClock: Integer;

    procedure SetClock (clock: Double);
    procedure SetVolume (gain: Byte);

    procedure WriteReg (index, v: byte);
    function ReadReg (index: Byte): Byte;

  public
    constructor Create ();

    procedure Init (clock: Double; pSampleRate: Integer);
    procedure InitClock (clock: Double); // 44100

    procedure Reset ();
    procedure Update ();
    procedure RenderSample ();

    property Registers [index: Byte]: Byte read ReadReg write WriteReg;
    property Frequency: Integer read fClock;

    // 0..ayMaxVolume
    property LeftChannel: Integer read fLeftChan;
    property RightChannel: Integer read fRightChan;
    property ChannelA: Integer read fChannelA;
    property ChannelB: Integer read fChannelB;
    property ChannelC: Integer read fChannelC;
  end;


implementation


const
  ayStep = 32768; // fixed-point multiplier
  maxOutput = 63;

  // AY register ID's
  ayRegAFine = 0;
  ayRegACoarse = 1;
  ayRegBFine = 2;
  ayRegBCoarse = 3;
  ayRegCFine = 4;
  ayRegCCoarse = 5;
  ayRegNoisePeriod = 6;
  ayRegEnable = 7;
  ayRegAVolume = 8;
  ayRegBVolume = 9;
  ayRegCVolume = 10;
  ayRegEFine = 11;
  ayRegECoarse = 12;
  ayRegEShape = 13;
  ayRegPortA = 14;
  ayRegPortB = 15;


constructor TAY8912.Create ();
begin
  Init(1773000, 44100);
end;

procedure TAY8912.Init (clock: Double; pSampleRate: Integer);
begin
  fAYPSG.sampleRate := pSampleRate;
  InitClock(clock);
end;

procedure TAY8912.InitClock (clock: Double); // 44100, 16 bit
begin
  SetClock(clock);
  SetVolume(12);
  Reset();
end;

procedure TAY8912.Reset ();
var
  f: Integer;
begin
  fAYPSG.registerLatch := 0;
  fAYPSG.outputA := 0;
  fAYPSG.outputB := 0;
  fAYPSG.outputC := 0;
  fAYPSG.outputN := $FF;
  fAYPSG.periodA := 0;
  fAYPSG.periodB := 0;
  fAYPSG.periodC := 0;
  fAYPSG.periodN := 0;
  fAYPSG.periodE := 0;
  fAYPSG.countA := 0;
  fAYPSG.countB := 0;
  fAYPSG.countC := 0;
  fAYPSG.countN := 0;
  fAYPSG.countE := 0;
  fAYPSG.volA := 0;
  fAYPSG.volB := 0;
  fAYPSG.volC := 0;
  fAYPSG.volE := 0;
  fAYPSG.envelopeA := 0;
  fAYPSG.envelopeB := 0;
  fAYPSG.envelopeC := 0;
  fAYPSG.countEnv := 0;
  fAYPSG.hold := 0;
  fAYPSG.alternate := 0;
  fAYPSG.holding := 0;
  fAYPSG.attack := 0;

  Randomize();
  for f := 0 to 14{ayRegPortA} do WriteReg(f, 0);  //???15???
end;

procedure TAY8912.SetClock (clock: Double);
// the ayStep clock for the tone and noise generators is the chip clock
// divided by 8; for the envelope generator of the AY-3-8912, it is half
// that much (clock/16), but the envelope of the YM2149 goes twice as
// fast, therefore again clock/8.
// Here we calculate the number of aySteps which happen during one sample
// at the given sample rate. No. of events = sample rate / (clock/8).
// ayStep is a multiplier used to turn the fraction into a fixed point
// number.
var
  t1: Double;
begin
  fClock := trunc(clock);
  t1 := ayStep*fAYPSG.sampleRate*8.0;
  fAYPSG.updateStep := t1/clock;
  //WriteLn('clk: ', clock:20:6); //1773000
  //WriteLn('t1 : ', t1:20:6);    //11560550400
  //WriteLn('up : ', fAYPSG.updateStep:20:6);   //6520.332995
  //WriteLn('sec: ', fAYPSG.updateStep/8:20:6); //815.041624
end;


procedure TAY8912.SetVolume (gain: Byte);
// set output gain
//
// The gain is expressed in 0.2dB increments, e.g. a gain of 10 is an increase
// of 2dB. Note that the gain only affects sounds not playing at full volume,
// since the ones at full volume are already played at the maximum intensity
// allowed by the sound card.
// $00 is the default.
// $FF is the maximum allowed value.
var
  f: Integer;
  out1, out2: Double;
begin
  gain := gain and $FF;

  // increase max output basing on gain (0.2 dB per ayStep)
  out1 := maxOutput; out2 := maxOutput;

  while gain > 0 do
  begin
    out1 := out1*1.023292992; // = (10 ^ (0.2/20))
    out2 := out2*1.023292992;
    Dec(gain);
  end;

  // calculate the volume.voltage conversion table
  // The AY-3-8912 has 16 levels, in a logarithmic scale (3dB per ayStep)
  // The YM2149 still has 16 levels for the tone generators, but 32 for
  // the envelope generator (1.5dB per ayStep).
  for f := 31 downto 0 do
  begin
    // limit volume to avoid clipping
    if out2 > maxOutput then fAYPSG.volTable2[f] := maxOutput
    else fAYPSG.volTable2[f] := round(out2);

    out1 := out1/1.188502227; // .188502227 = 10 ^ (1.5/20) = 1.5dB
    out2 := out2/1.188502227; // .188502227
  end;
  fAYPSG.volTable2[63] := maxOutput;
end;

procedure TAY8912.WriteReg (index, v: Byte);
// A note about the period of tones, noise and envelope: for speed reasons,
// we count down from the period to 0, but careful studies of the chip
// output prove that it instead counts up from 0 until the counter becomes
// greater or equal to the period. This is an important difference when the
// program is rapidly changing the period to modulate the sound.
// to compensate for the difference, when the period is changed we adjust
// our internal counter.
// Also, note that period = 0 is the same as period = 1. This is mentioned
// in the YM2203 data sheets. However, this does NOT apply to the Envelope
// period. In that case, period = 0 is half as period = 1.
var
  old: Integer;
begin
  if index > 15 then exit;
  fAYPSG.regs[index] := v;

  // adjust internal variables if necessary
  case index of
    ayRegAFine, ayRegACoarse: //0,1
      begin
        fAYPSG.regs[ayRegACoarse] := fAYPSG.regs[ayRegACoarse] and $0F;
        old := fAYPSG.periodA;
        fAYPSG.periodA := round((fAYPSG.regs[ayRegAFine]+(256*fAYPSG.regs[ayRegACoarse]))*fAYPSG.updateStep);
        if fAYPSG.periodA = 0 then fAYPSG.periodA := round(fAYPSG.updateStep);
        fAYPSG.countA := fAYPSG.countA+(fAYPSG.periodA-old);
        if fAYPSG.countA <= 0 then fAYPSG.countA := 1;
      end;
    ayRegBFine, ayRegBCoarse: //2,3
      begin
        fAYPSG.regs[ayRegBCoarse] := fAYPSG.regs[ayRegBCoarse] and $0F;
        old := fAYPSG.periodB;
        fAYPSG.periodB := round((fAYPSG.regs[ayRegBFine]+(256*fAYPSG.regs[ayRegBCoarse]))* fAYPSG.updateStep);
        if fAYPSG.periodB = 0 then fAYPSG.periodB := round(fAYPSG.updateStep);
        fAYPSG.countB := fAYPSG.countB+fAYPSG.periodB-old;
        if fAYPSG.countB <= 0 then fAYPSG.countB := 1;
      end;
    ayRegCFine, ayRegCCoarse: //4,5
      begin
        fAYPSG.regs[ayRegCCoarse] := fAYPSG.regs[ayRegCCoarse] and $0F;
        old := fAYPSG.periodC;
        fAYPSG.periodC := round((fAYPSG.regs[ayRegCFine]+(256*fAYPSG.regs[ayRegCCoarse]))*fAYPSG.updateStep);
        if fAYPSG.periodC = 0 then fAYPSG.periodC := round(fAYPSG.updateStep);
        fAYPSG.countC := fAYPSG.countC+(fAYPSG.periodC-old);
        if fAYPSG.countC <= 0 then fAYPSG.countC := 1;
      end;
    ayRegNoisePeriod: //6
      begin
        fAYPSG.regs[ayRegNoisePeriod] := fAYPSG.regs[ayRegNoisePeriod] and $1F;
        old := fAYPSG.periodN;
        fAYPSG.periodN := round(fAYPSG.regs[ayRegNoisePeriod]*fAYPSG.updateStep);
        if fAYPSG.periodN = 0 then fAYPSG.periodN := round(fAYPSG.updateStep);
        fAYPSG.countN := fAYPSG.countN+(fAYPSG.periodN-old);
        if fAYPSG.countN <= 0 then fAYPSG.countN := 1;
      end;
    ayRegAVolume: //8
      begin
        fAYPSG.regs[ayRegAVolume] := fAYPSG.regs[ayRegAVolume] and $1F;
        fAYPSG.envelopeA := fAYPSG.regs[ayRegAVolume] and $10;
        if fAYPSG.envelopeA <> 0 then fAYPSG.volA := fAYPSG.volE
        else
        begin
          if fAYPSG.regs[ayRegAVolume] <> 0 then fAYPSG.volA := fAYPSG.volTable2[fAYPSG.regs[ayRegAVolume]*2+1]
          else fAYPSG.volA := fAYPSG.volTable2[0];
        end;
      end;
    ayRegBVolume: //9
      begin
        fAYPSG.regs[ayRegBVolume] := fAYPSG.regs[ayRegBVolume] and $1F;
        fAYPSG.envelopeB := fAYPSG.regs[ayRegBVolume] and $10;
        if fAYPSG.envelopeB <> 0 then fAYPSG.volB := fAYPSG.volE
        else
        begin
          if fAYPSG.regs[ayRegBVolume] <> 0 then fAYPSG.volB := fAYPSG.volTable2[fAYPSG.regs[ayRegBVolume]*2+1]
          else fAYPSG.volB := fAYPSG.volTable2[0];
        end;
      end;
    ayRegCVolume: //10
      begin
        fAYPSG.regs[ayRegCVolume] := fAYPSG.regs[ayRegCVolume] and $1F;
        fAYPSG.envelopeC := fAYPSG.regs[ayRegCVolume] and $10;
        if fAYPSG.envelopeC <> 0 then fAYPSG.volC := fAYPSG.volE
        else
        begin
          if fAYPSG.regs[ayRegCVolume] <> 0 then fAYPSG.volC := fAYPSG.volTable2[fAYPSG.regs[ayRegCVolume]*2+1]
          else fAYPSG.volC := fAYPSG.volTable2[0];
        end;
      end;
    ayRegEFine, ayRegECoarse: //11,12
      begin
        old := fAYPSG.periodE;
        fAYPSG.periodE := round(((fAYPSG.regs[ayRegEFine]+(256*fAYPSG.regs[ayRegECoarse])))*fAYPSG.updateStep);
        if fAYPSG.periodE = 0 then fAYPSG.periodE := round(fAYPSG.updateStep/2);
        fAYPSG.countE := fAYPSG.countE+(fAYPSG.periodE-old);
        if fAYPSG.countE <= 0 then fAYPSG.countE := 1;
      end;
    ayRegEShape: //13
      begin
        // envelope shapes:
        // C AtAlH
        // 0 0 x x \___
        // 0 1 x x /___
        // 1 0 0 0 \\\\
        // 1 0 0 1 \___
        // 1 0 1 0 \/\/
        //         ___
        // 1 0 1 1    \
        // 1 1 0 0 ////
        //          ___
        // 1 1 0 1 /
        // 1 1 1 0 /\/\
        // 1 1 1 1 /___
        // The envelope counter on the AY-3-8910 has 16 aySteps.
        // On the YM2149 it has twice the aySteps, happening twice as fast.
        // Since the end result is just a smoother curve, we always use
        // the YM2149 behaviour.
        if fAYPSG.regs[ayRegEShape] <> $FF then
        begin
          fAYPSG.regs[ayRegEShape] := fAYPSG.regs[ayRegEShape] and $0F;
          if (fAYPSG.regs[ayRegEShape] and $04) <> 0 then fAYPSG.attack := ayMaxVolume
          else fAYPSG.attack := $0;
          fAYPSG.hold := fAYPSG.regs[ayRegEShape] and $1;
          fAYPSG.alternate := fAYPSG.regs[ayRegEShape] and $2;
          fAYPSG.countE := fAYPSG.periodE;
          fAYPSG.countEnv := ayMaxVolume; // &h1f
          fAYPSG.holding := 0;
          fAYPSG.volE := fAYPSG.volTable2[fAYPSG.countEnv xor fAYPSG.attack];
          if fAYPSG.envelopeA <> 0 then fAYPSG.volA := fAYPSG.volE;
          if fAYPSG.envelopeB <> 0 then fAYPSG.volB := fAYPSG.volE;
          if fAYPSG.envelopeC <> 0 then fAYPSG.volC := fAYPSG.volE;
        end;
      end;
  end;
end;

function TAY8912.ReadReg (index: Byte): Byte;
begin
  if index > 15 then result := 0 else result := fAYPSG.regs[index];
end;

procedure TAY8912.Update ();
// The 8910 has three outputs, each output is the mix of one of the three
// tone generators and of the (single) noise generator. The two are mixed
// BEFORE going into the DAC. The formula to mix each channel is:
// (ToneOn | ToneDisable) & (NoiseOn | NoiseDisable).
// Note that this means that if both tone and noise are disabled, the output
// is 1, not 0, and can be modulated changing the volume.
// if the channels are disabled, set their output to 1, and increase the
// counter, if necessary, so they will not be inverted during this update.
// Setting the output to 1 is necessary because a disabled channel is locked
// into the ON state (see above); and it has no effect if the volume is 0.
// if the volume is 0, increase the counter, but don't touch the output.
// note that I do count += bufferLength, NOT count = bufferLength + 1.
// You might think it's the same since the volume is 0, but doing the latter
// could cause interferencies when the program is rapidly modulating the volume.
// for the noise channel we must not touch outputN - it's also not necessary
// since we use ayOutNoise.
var
  bufferLength, bufLenMulayStep: Integer;
begin
  bufferLength := 400;
  bufLenMulayStep := bufferLength*ayStep;

  if (fAYPSG.regs[ayRegEnable] and $01) <> 0 then
  begin
    if fAYPSG.countA <= bufLenMulayStep then Inc(fAYPSG.countA, bufLenMulayStep);
    fAYPSG.outputA := 1;
  end
  else if fAYPSG.regs[ayRegAVolume] = 0 then
  begin
    if fAYPSG.countA <= bufLenMulayStep then Inc(fAYPSG.countA, bufLenMulayStep);
  end;

  if (fAYPSG.regs[ayRegEnable] and $02) <> 0 then
  begin
    if fAYPSG.countB <= bufLenMulayStep then Inc(fAYPSG.countB, bufLenMulayStep);
    fAYPSG.outputB := 1;
  end
  else if fAYPSG.regs[ayRegBVolume] = 0 then
  begin
    if fAYPSG.countB <= bufLenMulayStep then Inc(fAYPSG.countB, bufLenMulayStep);
  end;

  if (fAYPSG.regs[ayRegEnable] and $04) <> 0 then
  begin
    if fAYPSG.countC <= bufLenMulayStep then Inc(fAYPSG.countC, bufLenMulayStep);
    fAYPSG.outputC := 1;
  end
  else if fAYPSG.regs[ayRegCVolume] = 0 then
  begin
    if fAYPSG.countC <= bufLenMulayStep then Inc(fAYPSG.countC, bufLenMulayStep);
  end;

  if (fAYPSG.regs[ayRegEnable] and $38) = $38 then // all off
  begin
    if fAYPSG.countN <= bufLenMulayStep then Inc(fAYPSG.countN, bufLenMulayStep);
  end;

  ayOutNoise := (fAYPSG.outputN or fAYPSG.regs[ayRegEnable]);
end;

procedure TAY8912.RenderSample ();
begin
  // volA, volB and volC keep track of how long each square wave stays
  // in the 1 position during the sample period.
  volA := 0; volB := 0; volC := 0;

  ayLeft := ayStep;
  repeat
    ayNextEvent := 0;

    if fAYPSG.countN < ayLeft then ayNextEvent := fAYPSG.countN else ayNextEvent := ayLeft;
    if (ayOutNoise and $08) <> 0 then
    begin
      if fAYPSG.outputA = 1 then Inc(volA, fAYPSG.countA);
      Dec(fAYPSG.countA, ayNextEvent);

      // periodA is the half period of the square wave. Here, in each
      // loop I add periodA twice, so that at the end of the loop the
      // square wave is in the same status (0 or 1) it was at the start.
      // volA is also incremented by periodA, since the wave has been 1
      // exactly half of the time, regardless of the initial position.
      // if we exit the loop in the middle, outputA has to be inverted
      // and volA incremented only if the exit status of the square
      // wave is 1.
      while fAYPSG.countA <= 0 do
      begin
       Inc(fAYPSG.countA, fAYPSG.periodA);
       if fAYPSG.countA > 0 then
       begin
         if (fAYPSG.regs[ayRegEnable] and 1) = 0 then fAYPSG.outputA := fAYPSG.outputA xor 1;
         if fAYPSG.outputA <> 0 then Inc(volA, fAYPSG.periodA);
         break;
       end;
       Inc(fAYPSG.countA, fAYPSG.periodA);
       Inc(volA, fAYPSG.periodA);
      end;
      if fAYPSG.outputA = 1 then Dec(volA, fAYPSG.countA);
    end
    else
    begin
      Dec(fAYPSG.countA, ayNextEvent);
      while fAYPSG.countA <= 0 do
      begin
        Inc(fAYPSG.countA, fAYPSG.periodA);
        if fAYPSG.countA > 0 then
        begin
          fAYPSG.outputA := fAYPSG.outputA xor 1;
          break;
        end;
        Inc(fAYPSG.countA, fAYPSG.periodA);
      end;
    end;

    if (ayOutNoise and $10) <> 0 then
    begin
      if fAYPSG.outputB = 1 then Inc(volB, fAYPSG.countB);
      Dec(fAYPSG.countB, ayNextEvent);
      while fAYPSG.countB <= 0 do
      begin
        Inc(fAYPSG.countB, fAYPSG.periodB);
        if fAYPSG.countB > 0 then
        begin
          if (fAYPSG.regs[ayRegEnable] and 2) = 0 then fAYPSG.outputB := fAYPSG.outputB xor 1;
          if fAYPSG.outputB <> 0 then Inc(volB, fAYPSG.periodB);
          break;
        end;
        Inc(fAYPSG.countB, fAYPSG.periodB);
        Inc(volB, fAYPSG.periodB);
      end;
      if fAYPSG.outputB = 1 then Dec(volB, fAYPSG.countB);
    end
    else
    begin
      Dec(fAYPSG.countB, ayNextEvent);
      while fAYPSG.countB <= 0 do
      begin
        Inc(fAYPSG.countB, fAYPSG.periodB);
        if fAYPSG.countB > 0 then
        begin
          fAYPSG.outputB := fAYPSG.outputB xor 1;
          break;
        end;
        Inc(fAYPSG.countB, fAYPSG.periodB);
      end;
    end;

    if (ayOutNoise and $20) <> 0 then
    begin
      if fAYPSG.outputC = 1 then Inc(volC, fAYPSG.countC);
      Dec(fAYPSG.countC, ayNextEvent);
      while fAYPSG.countC <= 0 do
      begin
        Inc(fAYPSG.countC, fAYPSG.periodC);
        if fAYPSG.countC > 0 then
        begin
          if (fAYPSG.regs[ayRegEnable] and 4) = 0 then fAYPSG.outputC := fAYPSG.outputC xor 1;
          if fAYPSG.outputC <> 0 then Inc(volC, fAYPSG.periodC);
          break;
        end;
        Inc(fAYPSG.countC, fAYPSG.periodC);
        Inc(volC, fAYPSG.periodC);
      end;
      if fAYPSG.outputC = 1 then Dec(volC, fAYPSG.countC);
    end
    else
    begin
      Dec(fAYPSG.countC, ayNextEvent);
      while fAYPSG.countC <= 0 do
      begin
        Inc(fAYPSG.countC, fAYPSG.periodC);
        if fAYPSG.countC > 0 then
        begin
          fAYPSG.outputC := fAYPSG.outputC xor 1;
          break;
        end;
        Inc(fAYPSG.countC, fAYPSG.periodC);
      end;
    end;

    Dec(fAYPSG.countN, ayNextEvent);
    if fAYPSG.countN <= 0 then
    begin
      //Is noise output going to change?
      fAYPSG.outputN := round(random(510));
      ayOutNoise := (fAYPSG.outputN or fAYPSG.regs[ayRegEnable]);
      Inc(fAYPSG.countN, fAYPSG.periodN);
    end;

    Dec(ayLeft, ayNextEvent);
  until ayLeft <= 0;

  if fAYPSG.holding = 0 then
  begin
    Dec(fAYPSG.countE, ayStep);
    if fAYPSG.countE <= 0 then
    begin
      repeat
        Dec(fAYPSG.countEnv);
        Inc(fAYPSG.countE, fAYPSG.periodE);
      until fAYPSG.countE > 0;

      //check envelope current position
      if fAYPSG.countEnv < 0 then
      begin
        if fAYPSG.hold <> 0 then
        begin
          if fAYPSG.alternate <> 0 then
          begin
            fAYPSG.attack := fAYPSG.attack xor ayMaxVolume; //$1f
          end;
          fAYPSG.holding := 1;
          fAYPSG.countEnv := 0;
        end
        else
        begin
          //if countEnv has looped an odd number of times (usually 1),
          //invert the output.
          if (fAYPSG.alternate <> 0) and ((fAYPSG.countEnv and $20) <> 0) then
          begin
            fAYPSG.attack := fAYPSG.attack xor ayMaxVolume; //$1f
          end;
          fAYPSG.countEnv := fAYPSG.countEnv and ayMaxVolume; //$1f
        end;
      end;
      fAYPSG.volE := fAYPSG.volTable2[fAYPSG.countEnv xor fAYPSG.attack];

      //reload volume
      if fAYPSG.envelopeA <> 0 then fAYPSG.volA := fAYPSG.volE;
      if fAYPSG.envelopeB <> 0 then fAYPSG.volB := fAYPSG.volE;
      if fAYPSG.envelopeC <> 0 then fAYPSG.volC := fAYPSG.volE;
    end;
  end;

  lOut0 := (volA*fAYPSG.volA) div 65535;
  lOut1 := (volB*fAYPSG.volB) div 65535;
  lOut2 := (volC*fAYPSG.volC) div 65535;
  fChannelA := lOut0; fChannelB := lOut1; fChannelC := lOut2;

  // ABC mode
  // FIXME: write code for mode selection!
  lOut0 := (lOut0+lOut2 div 2); // 0..47
  lOut1 := (lOut1+lOut2 div 2); // 0..47

  fLeftChan := lOut0; fRightChan := lOut1;
end;


end.
